from .saveable import Saveable
from .animsprite import AnimSprite
from .input import InputType, InputDefault, EVENT_TYPE_START_POS
from .modebase import ModeBase
from .fontwrap import FontWrap, getDefaultFontWrap
from . import load
from . import utility
from .game import initGame, getGame



import sys
import os
import math

import pygame

from . import config
from . import utility
from .modebase import ModeBase


_screenshot_directory: str
screen_size: tuple[int, int]
_title: str
_window_icon: pygame.Surface | None
_monitor_res: tuple[int, int]
_upscale_max: int
_windowed_flags: int
_fullscreen_flags: int
is_fullscreen: bool
upscale: int
_disp_res: tuple[int, int]
screen: pygame.Surface | None = None
_fullscreen_offset: tuple[int, int] | None
_full_screen: pygame.Surface | None
_disp_screen: pygame.Surface


def init(
    screenshot_directory: str,
    screen_size_in: tuple[int, int],
    title: str,
    window_icon: str | None
):
    global _screenshot_directory
    global screen_size
    global _title
    global _window_icon
    global is_fullscreen
    global upscale
    global screen
    global _fullscreen_offset
    global _full_screen
    if screen:
        raise RuntimeError("error: screen is already set")
    _screenshot_directory = screenshot_directory
    screen_size = screen_size_in
    _title = title
    _window_icon = None
    if window_icon:
        _window_icon = pygame.image.load(window_icon)
    _setupDisplay()
    is_fullscreen = config.get(config.FULLSCREEN)
    upscale = config.get(config.SCREEN_SCALE)
    if upscale == 0:
        upscale = math.ceil(_upscale_max / 2)
    upscale = utility.clamp(upscale, 0, _upscale_max)
    _scaleDisp()
    screen = pygame.Surface(screen_size)
    _fullscreen_offset = None
    _full_screen = None
    if is_fullscreen:
        _setFullscreen()
    else:
        _setWindowed()
    screen = screen.convert()
    config.update(config.SCREEN_SCALE, upscale)


def _setupDisplay():
    global _monitor_res
    global _upscale_max
    global _windowed_flags
    global _fullscreen_flags
    pygame.display.set_caption(_title)
    if _window_icon:
        pygame.display.set_icon(_window_icon)
    display_info = pygame.display.Info()
    _monitor_res = (
        display_info.current_w,
        display_info.current_h,
    )
    _upscale_max = min(
        _monitor_res[0] // screen_size[0],
        _monitor_res[1] // screen_size[1]
    )
    max_disp_res = (
        screen_size[0] * _upscale_max,
        screen_size[1] * _upscale_max,
    )
    _windowed_flags = 0
    if pygame.display.mode_ok(max_disp_res, pygame.DOUBLEBUF):
        _windowed_flags = pygame.DOUBLEBUF
    _fullscreen_flags = 0
    if sys.platform != "win32":
        _fullscreen_flags = pygame.NOFRAME
    elif pygame.display.mode_ok(_monitor_res, pygame.FULLSCREEN | pygame.DOUBLEBUF | pygame.HWSURFACE):
        _fullscreen_flags = pygame.FULLSCREEN | pygame.DOUBLEBUF | pygame.HWSURFACE
    elif pygame.display.mode_ok(_monitor_res, pygame.FULLSCREEN | pygame.DOUBLEBUF):
        _fullscreen_flags = pygame.FULLSCREEN | pygame.DOUBLEBUF
    elif pygame.display.mode_ok(_monitor_res, pygame.FULLSCREEN):
        _fullscreen_flags = pygame.FULLSCREEN


def changeScale(scale_change: int):
    new_scale = upscale + scale_change
    if new_scale < 1 or new_scale > _upscale_max:
        return
    _alterScale(new_scale)


def setScale(target_scale: int):
    new_scale = min(target_scale, _upscale_max)
    if new_scale == upscale:
        return
    _alterScale(new_scale)


def _alterScale(new_scale: int):
    global upscale
    global screen
    upscale = new_scale
    _scaleDisp()
    if is_fullscreen:
        _setFullscreen()
    else:
        pygame.display.quit()
        pygame.display.init()
        _setupDisplay()
        _setWindowed()
    screen = screen.convert()
    config.update(config.SCREEN_SCALE, upscale)


def _scaleDisp():
    global _disp_res
    _disp_res = (
        screen_size[0] * upscale,
        screen_size[1] * upscale,
    )


def toggleFullscreen():
    global is_fullscreen
    global screen
    is_fullscreen = not is_fullscreen
    pygame.display.quit()
    pygame.display.init()
    _setupDisplay()
    if is_fullscreen:
        _setFullscreen()
    else:
        _setWindowed()
    screen = screen.convert()
    config.update(config.FULLSCREEN, is_fullscreen)


def _setWindowed():
    global _fullscreen_offset
    global _full_screen
    global _disp_screen
    # center window
    x = (_monitor_res[0] - _disp_res[0]) // 2
    y = (_monitor_res[1] - _disp_res[1]) // 2
    os.environ['SDL_VIDEO_WINDOW_POS'] = f'{x},{y}'
    _fullscreen_offset = None
    _full_screen = None
    _disp_screen = pygame.display.set_mode(
        _disp_res,
        _windowed_flags
    )


def _setFullscreen():
    global _fullscreen_offset
    global _full_screen
    global _disp_screen
    _fullscreen_offset = (
        (_monitor_res[0] - _disp_res[0]) // 2,
        (_monitor_res[1] - _disp_res[1]) // 2,
    )
    if _full_screen is None:
        os.environ['SDL_VIDEO_WINDOW_POS'] = '0,0'
        _full_screen = pygame.display.set_mode(
            _monitor_res,
            _fullscreen_flags
        )
    else:
        _full_screen.fill((0, 0, 0))
    _disp_screen = pygame.Surface(_disp_res).convert()


def scaleDraw():
    """Scale screen onto display surface, then flip the display."""
    pygame.transform.scale(screen, _disp_res, _disp_screen)
    if is_fullscreen:
        _full_screen.blit(_disp_screen, _fullscreen_offset)
    pygame.display.flip()


def scaleMouseInput(event: pygame.event.Event):
    """Scale mouse position for events in terms of the screen (as opposed to the display surface)."""
    if event.type in (pygame.MOUSEMOTION, pygame.MOUSEBUTTONUP, pygame.MOUSEBUTTONDOWN):
        if is_fullscreen:
            event_dict = {
                'pos': (
                    (event.pos[0] - _fullscreen_offset[0]) // upscale,
                    (event.pos[1] - _fullscreen_offset[1]) // upscale,
                )
            }
        else:
            event_dict = {
                'pos': (
                    event.pos[0] // upscale,
                    event.pos[1] // upscale,
                )
            }
        if event.type == pygame.MOUSEMOTION:
            event_dict['rel'] = (
                event.rel[0] // upscale,
                event.rel[1] // upscale,
            )
            event_dict['buttons'] = event.buttons
        else:
            event_dict['button'] = event.button
        return pygame.event.Event(event.type, event_dict)
    return event


def isInScreen(pos: tuple[int, int]):
    return (
        0 <= pos[0] < screen_size[0]
        and 0 <= pos[1] < screen_size[1]
    )


def getBlurredScreen(mode: ModeBase):
    result = pygame.Surface(screen_size).convert()
    mode.draw(result)
    result = pygame.transform.smoothscale(
        result,
        (screen_size[0] * 4 // 5, screen_size[1] * 4 // 5)
    )
    result = pygame.transform.smoothscale(
        result,
        screen_size
    )
    return result


def getPositionalChannelMix(x: int | float):
    pos = utility.clamp(x / screen_size[0], 0, 1)
    channel_l = _boundChannelVolume(utility.cosCurve(pos))
    channel_r = _boundChannelVolume(utility.sinCurve(pos))
    return channel_l, channel_r


def _boundChannelVolume(volume: float):
    return .2 + (volume * .8)


def takeScreenshot():
    try:
        os.mkdir(_screenshot_directory)
    except FileExistsError:
        pass
    file_name = f'{utility.getDateTimeFileName()}.png'
    file_path = os.path.join(_screenshot_directory, file_name)
    pygame.image.save(screen, file_path)



import os
from types import ModuleType
from collections.abc import Iterable

import pygame
pygame.init()

from . import display
from . import input
from . import fontwrap
from .modebase import ModeBase
from .modegamemenu import ModeGameMenu
from .modegamemenu import ModeGameMenuTop
from .saveable import Saveable
from . import config
from . import save


class _Game(object):
    __slots__ = (
        'start_mode_cls',
        'state_cls',
        '_clock',
        '_max_framerate',
        '_joysticks',
        '_is_first_loop',
        '_current_mode',
        'state',
        'running',
    )

    def __init__(
        self,
        mode_module: ModuleType,
        start_mode_cls: type[ModeBase],
        state_cls: type[Saveable],
        src_directory: str,
        screen_size: tuple[int, int],
        title: str,
        window_icon: str | None,
        max_players: int,
        event_names: tuple[str],
        input_defaults: tuple[input.InputDefault],
        font_location: str | None,
        font_size: int,
        font_height: int,
        font_antialias: bool
    ):
        self.start_mode_cls = start_mode_cls
        self.state_cls = state_cls
        config.init(
            os.path.join(src_directory, 'config.ini')
        )
        save.init(
            os.path.join(src_directory, 'saves'),
            mode_module
        )
        display.init(
            os.path.join(src_directory, 'screenshots'),
            screen_size,
            title,
            window_icon
        )
        input.init(
            os.path.join(src_directory, 'input.cfg'),
            max_players,
            event_names,
            input_defaults
        )
        if font_location:
            font = pygame.font.Font(font_location, font_size)
        else:
            font = pygame.font.SysFont(None, font_size)
        fontwrap.init(font, font_height, font_antialias)
        self._clock = pygame.time.Clock()
        self._max_framerate = config.get(config.MAX_FRAMERATE)
        self._joysticks = [
            pygame.joystick.Joystick(i)
            for i
            in range(pygame.joystick.get_count())
        ]
        self._is_first_loop = True
        self._current_mode: ModeBase | None = None
        self.state: Saveable | None = None
        self.running = False

    def start(self):
        """Start the game, must be called before run()."""
        self._current_mode = self.start_mode_cls()
        self.state = self.state_cls()
        self.running = True

    def run(self):
        """Run the game, and check if the game needs to end."""
        if not self._current_mode:
            raise RuntimeError("error: self._current_mode is not set")
        events = self._filterInput(pygame.event.get())
        events = input.takeEvents(events)
        input_frame = input.getInputFrame()
        if input_frame.wasPlayerInputPressed(0, input.TYPE_SCREENSHOT):
            display.takeScreenshot()
        if any(map(self._isPauseEvent, events)) or input_frame.wasInputPressed(input.TYPE_PAUSE):
            # if already in pause menu no need to do this stuff
            if not isinstance(self._current_mode, ModeGameMenu):
                self._current_mode = ModeGameMenuTop(self._current_mode)
                input.startNewMode()
                input_frame = input.getInputFrame()
                pygame.mixer.music.pause()
                pygame.mixer.pause()
                events = []
        self._current_mode.input(events, input_frame)
        for i in range(self._getTime()):
            self._current_mode.update(1)
        self._current_mode.draw(display.screen)
        display.scaleDraw()
        if self._current_mode.next_mode is not None:
            if isinstance(self._current_mode, ModeGameMenu) \
                    and not isinstance(self._current_mode.next_mode, ModeGameMenu):
                pygame.mixer.music.unpause()
                pygame.mixer.unpause()
            self._current_mode.cleanup()
            self._current_mode = self._current_mode.next_mode
            input.startNewMode()
        self._is_first_loop = False
        if not self.running:
            config.save()
            input.save()
            self._current_mode = None
            self.state = None
            pygame.quit()
        return self.running

    def _filterInput(self, events: Iterable[pygame.event.Event]):
        """Take care of input that game modes should not take care of."""
        events = map(display.scaleMouseInput, events)
        events = filter(self._filterEvent, events)
        return list(events)

    def _filterEvent(self, event: pygame.event.Event):
        """If event should be handled before all others, handle it and return False, otherwise return True.
        As an example, game-ending or display-changing events should be handled before all others.
        Also filter out bad mouse events here.
        """
        match event.type:
            case pygame.MOUSEMOTION | pygame.MOUSEBUTTONUP | pygame.MOUSEBUTTONDOWN:
                return display.isInScreen(event.pos)
            case pygame.JOYDEVICEREMOVED:
                self._joysticks = [
                    joystick
                    for joystick
                    in self._joysticks
                    if joystick.get_instance_id() != event.instance_id
                ]
                return False
            case pygame.JOYDEVICEADDED:
                self._joysticks = [
                    pygame.joystick.Joystick(i)
                    for i
                    in range(pygame.joystick.get_count())
                ]
                return False
        return True

    def _isPauseEvent(self, event: pygame.event.Event):
        return event.type in (pygame.QUIT, pygame.WINDOWFOCUSLOST, pygame.WINDOWMINIMIZED) \
            or (event.type == pygame.WINDOWMOVED and not self._is_first_loop)

    def _getTime(self):
        return self._clock.tick(self._max_framerate)


_game: _Game | None = None


def initGame(
    mode_module: ModuleType,
    start_mode_cls: type[ModeBase],
    state_cls: type[Saveable],
    src_directory: str,
    screen_size: tuple[int, int],
    title: str,
    window_icon: str | None,
    max_players: int,
    event_names: tuple[str],
    input_defaults: tuple[input.InputDefault],
    font_location: str | None,
    font_size: int,
    font_height: int,
    font_antialias: bool
):
    """Loads up the game and prepares it for running.
    Arguments:
    mode_module: the module holding all modes for your game
    start_mode_cls: the class for the first mode
    state_cls: the class for holding general game state
    src_directory: directory of the program
    screen_size: size of the virtual screen
    title: title of the game (for titlebar)
    window_icon: location of icon of the game (for titlebar)
    max_players: maximum number of players the game supports
    event_names: names for virtual inputs that button / axis inputs map to
    input_defaults: default input mappings
    font_location: location of default font for the game
    font_size: default font size
    font_height: default font height
    font_antialias: default font antialias
    """
    global _game
    if _game:
        raise RuntimeError("error: _game is already set")
    _game = _Game(
        mode_module,
        start_mode_cls,
        state_cls,
        src_directory,
        screen_size,
        title,
        window_icon,
        max_players,
        event_names,
        input_defaults,
        font_location,
        font_size,
        font_height,
        font_antialias
    )
    _game.start()
    return _game


def getGame():
    return _game



import unittest

import pygame

import jovialengine.display as display
import jovialengine.game as game


class GameForTest(game._Game):
    def __init__(self):
        self.running = False
        self._is_first_loop = False
        self._joysticks = []


class TestGame(unittest.TestCase):
    game_for_test: game._Game = None

    @classmethod
    def setUpClass(cls):
        cls.game_for_test = GameForTest()
        display.is_fullscreen = False
        display._fullscreen_offset = None
        display.upscale = 3
        display.screen_size = (320, 240)

    def test__isPauseEvent_QUIT(self):
        # Arrange
        event_dict = {
        }
        event = pygame.event.Event(pygame.QUIT, event_dict)
        # Act
        result = self.game_for_test._isPauseEvent(event)
        # Assert
        self.assertTrue(result)

    def test__isPauseEvent_WINDOWFOCUSLOST(self):
        # Arrange
        event_dict = {
        }
        event = pygame.event.Event(pygame.WINDOWFOCUSLOST, event_dict)
        # Act
        result = self.game_for_test._isPauseEvent(event)
        # Assert
        self.assertTrue(result)

    def test__isPauseEvent_WINDOWMINIMIZED(self):
        # Arrange
        event_dict = {
        }
        event = pygame.event.Event(pygame.WINDOWMINIMIZED, event_dict)
        # Act
        result = self.game_for_test._isPauseEvent(event)
        # Assert
        self.assertTrue(result)

    def test__isPauseEvent_WINDOWMOVED(self):
        # Arrange
        event_dict = {
        }
        event = pygame.event.Event(pygame.WINDOWMOVED, event_dict)
        # Act
        result = self.game_for_test._isPauseEvent(event)
        # Assert
        self.assertTrue(result)

    def test__filterEvent_KEYDOWN_a(self):
        # Arrange
        event_dict = {
            'key': pygame.K_a,
        }
        event = pygame.event.Event(pygame.KEYDOWN, event_dict)
        # Act
        result = self.game_for_test._filterEvent(event)
        # Assert
        self.assertTrue(result)

    def test__filterEvent_MOUSEBUTTONDOWN_IN(self):
        # Arrange
        event_dict = {
            'pos': (1, 1),
            'button': 1
        }
        event = pygame.event.Event(pygame.MOUSEBUTTONDOWN, event_dict)
        # Act
        result = self.game_for_test._filterEvent(event)
        # Assert
        self.assertTrue(result)

    def test__filterEvent_MOUSEBUTTONDOWN_OUT(self):
        # Arrange
        event_dict = {
            'pos': (-1, -1),
            'button': 1
        }
        event = pygame.event.Event(pygame.MOUSEBUTTONDOWN, event_dict)
        # Act
        result = self.game_for_test._filterEvent(event)
        # Assert
        self.assertFalse(result)


if __name__ == '__main__':
    unittest.main()



import unittest
import os

import jovialengine.save as save

import mode


class TestSave(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        save.init(
            os.path.join('.', 'saves'),
            mode
        )

    def test0_getAllFromFiles(self):
        # Act
        saves = save.Save.getAllFromFiles()
        # Assert
        self.assertEqual(len(saves), 1)
        self.assertEqual(saves[0].save_name, "a")
        self.assertEqual(saves[0]._mode_name, "ModeTest")
        self.assertEqual(saves[0]._mode_data, 1)
        self.assertEqual(saves[0]._shared_data, "test")

    def test1_willOverwrite(self):
        # Arrange
        collide_name = "a"
        noncollide_name = "c"
        # Act
        collide_result = save.Save.willOverwrite(collide_name)
        noncollide_result = save.Save.willOverwrite(noncollide_name)
        # Assert
        self.assertEqual(collide_result, True)
        self.assertEqual(noncollide_result, False)

    def test2_save_existing(self):
        # Arrange
        saves = save.Save.getAllFromFiles()
        test_save = saves[0]
        test_save._mode_data = 2
        # Act
        test_save.save()
        saves = save.Save.getAllFromFiles()
        # Assert
        self.assertEqual(len(saves), 1)
        self.assertEqual(saves[0].save_name, "a")
        self.assertEqual(saves[0]._mode_name, "ModeTest")
        self.assertEqual(saves[0]._mode_data, 2)
        self.assertEqual(saves[0]._shared_data, "test")

    def test3_save_new(self):
        # Arrange
        saves = save.Save.getAllFromFiles()
        test_save = saves[0]
        test_save.save_name = "b"
        test_save._mode_data = 2
        # Act
        test_save.save()
        saves = save.Save.getAllFromFiles()
        # Assert
        self.assertEqual(len(saves), 2)
        self.assertEqual(saves[1].save_name, "b")
        self.assertEqual(saves[1]._mode_name, "ModeTest")
        self.assertEqual(saves[1]._mode_data, 2)
        self.assertEqual(saves[1]._shared_data, "test")

    def test4_save_delete(self):
        # Arrange
        saves = save.Save.getAllFromFiles()
        # Act
        for test_save in saves:
            test_save.delete()
        saves = save.Save.getAllFromFiles()
        # Assert
        self.assertEqual(len(saves), 0)


if __name__ == '__main__':
    unittest.main()



import abc
import enum
import string

import pygame

from . import game
from . import display
from . import input
from . import utility
from .fontwrap import getDefaultFontWrap
from .modebase import ModeBase
from .save import Save
from .saveable import Saveable


class MenuAction(enum.Enum):
    NOTHING = enum.auto()
    LEFT = enum.auto()
    RIGHT = enum.auto()
    UP = enum.auto()
    DOWN = enum.auto()
    CONFIRM = enum.auto()
    REJECT = enum.auto()
    QUIT = enum.auto()


class ModeGameMenu(ModeBase, abc.ABC):
    _TEXT_COLOR = (255, 255, 255)
    _BACKGROUND_COLOR = (0, 0, 0)
    _MENU_CHAR_WIDTH = 40
    _SHARED_DISP_TEXT = "Options:\nESC) Go Back\n"

    __slots__ = (
        '_MENU_WIDTH',
        '_previous_mode',
        '_last_disp_text',
        '_menu_surface',
        '_previous_hat',
    )

    def __init__(self, previous_mode: ModeBase, old_screen: pygame.Surface | None = None):
        self._init(display.screen_size)
        self._MENU_WIDTH = getDefaultFontWrap().font.size('_' * self._MENU_CHAR_WIDTH)[0] + 1
        self._previous_mode = previous_mode
        if old_screen is None:
            old_screen = self._getOldScreen()
        self._background = old_screen
        self._last_disp_text: str | None = None
        self._menu_surface: pygame.Surface
        self._previous_hat = (0, 0)

    def _getOldScreen(self):
        return display.getBlurredScreen(self._previous_mode)

    def _getAction(self, event: pygame.event.Event):
        match event.type:
            case pygame.QUIT:
                return MenuAction.QUIT
            case pygame.JOYHATMOTION:
                result = MenuAction.NOTHING
                if event.value[0] == -1 != self._previous_hat[0]:
                    result = MenuAction.LEFT
                elif event.value[0] == 1 != self._previous_hat[0]:
                    result = MenuAction.RIGHT
                elif event.value[1] == 1 != self._previous_hat[1]:
                    result = MenuAction.UP
                elif event.value[1] == -1 != self._previous_hat[1]:
                    result = MenuAction.DOWN
                self._previous_hat = event.value
                return result
            case pygame.JOYBUTTONDOWN:
                if event.button == 0:
                    return MenuAction.CONFIRM
                else:
                    return MenuAction.REJECT
            case pygame.KEYDOWN:
                match event.key:
                    case pygame.K_LEFT:
                        return MenuAction.LEFT
                    case pygame.K_RIGHT:
                        return MenuAction.RIGHT
                    case pygame.K_UP:
                        return MenuAction.UP
                    case pygame.K_DOWN:
                        return MenuAction.DOWN
                    case pygame.K_RETURN:
                        return MenuAction.CONFIRM
                    case pygame.K_ESCAPE:
                        return MenuAction.REJECT
        return MenuAction.NOTHING

    @staticmethod
    def _getSelectedChar(is_selected: bool):
        return ">" if is_selected else "_"

    def _drawTextAlways(self, disp_text: str):
        self._last_disp_text = disp_text
        self._menu_surface = getDefaultFontWrap().renderInside(
            self._MENU_WIDTH,
            disp_text,
            self._TEXT_COLOR,
            self._BACKGROUND_COLOR
        )
        self._menu_surface.set_alpha(235)

    def _drawText(self, disp_text: str):
        if self._last_disp_text != disp_text:
            self._drawTextAlways(disp_text)


class ModeGameMenuList(ModeGameMenu, abc.ABC):
    __slots__ = (
        '_index',
    )

    def __init__(self, previous_mode, old_screen):
        super().__init__(previous_mode, old_screen)
        self._index = 0

    @abc.abstractmethod
    def _getOptionsLength(self) -> int:
        raise NotImplementedError(
            type(self).__name__ + "._getOptionsLength(self)"
        )

    @abc.abstractmethod
    def _getOptionName(self, index: int) -> str:
        raise NotImplementedError(
            type(self).__name__ + "._getOptionName(self, index)"
        )

    def _getOptionsText(self) -> str:
        text = ""
        for i in range(-1, 2):
            text += "\n"
            this_index = self._index + i
            text += ">" if i == 0 else "_"
            if 0 <= this_index < self._getOptionsLength():
                text += self._getOptionName(this_index)
        return text


class ModeGameMenuTop(ModeGameMenu):
    _OPTIONS = [
        "Save",
        "Load",
        "Options",
        "Edit Controls",
        "Reset Controls",
        "Restart",
        "Quit",
    ]

    __slots__ = (
        '_selected',
    )

    def __init__(self, previous_mode, old_screen=None):
        super().__init__(previous_mode, old_screen)
        self._selected = 0

    def _inputEvent(self, event):
        match self._getAction(event):
            case MenuAction.QUIT:
                game.getGame().running = False
            case MenuAction.REJECT:
                self.next_mode = self._previous_mode
            case MenuAction.UP | MenuAction.LEFT:
                self._selected -= 1
            case MenuAction.DOWN | MenuAction.RIGHT:
                self._selected += 1
            case MenuAction.CONFIRM:
                match self._selected:
                    case 0:
                        self.next_mode = ModeGameMenuSave(self._previous_mode, self._background)
                    case 1:
                        self.next_mode = ModeGameMenuLoad(self._previous_mode, self._background)
                    case 2:
                        pressed_return = event.type == pygame.KEYDOWN and event.key == pygame.K_RETURN
                        self.next_mode = ModeGameMenuOptions(self._previous_mode, self._background, pressed_return)
                    case 3:
                        self.next_mode = ModeGameMenuControls(self._previous_mode, self._background)
                    case 4:
                        input.resetDefaultMapping()
                    case 5:
                        self._stopMixer()
                        game.getGame().state = game.getGame().state_cls()
                        self._previous_mode = game.getGame().start_mode_cls()
                        pygame.mixer.music.pause()
                        pygame.mixer.pause()
                        self._background = self._getOldScreen()
                        self._last_disp_text = None
                    case 6:
                        game.getGame().running = False
        self._selected = utility.clamp(self._selected, 0, len(self._OPTIONS) - 1)

    def _drawPreSprites(self, screen):
        disp_text = self._SHARED_DISP_TEXT
        disp_text += "ARROW KEYS + ENTER)"
        for index, option in enumerate(self._OPTIONS):
            disp_text += "\n"
            disp_text += self._getSelectedChar(self._selected == index)
            disp_text += option
        self._drawText(disp_text)
        screen.blit(self._menu_surface, (0, 0))


class ModeGameMenuSave(ModeGameMenu):
    _CONTROLLER_CHARS = string.ascii_lowercase + string.digits
    _CURSOR_TIME = 500

    __slots__ = (
        '_save_name',
        '_cursor_position',
        '_confirm_overwrite',
        '_save_success',
        '_cursor_switch',
        '_cursor_timer',
    )

    def __init__(self, previous_mode, old_screen):
        super().__init__(previous_mode, old_screen)
        self._save_name = ''
        self._resetCursorBlink()
        self._cursor_position = 0
        self._confirm_overwrite = False
        self._save_success = None

    def _resetCursorBlink(self):
        self._cursor_switch = True
        self._cursor_timer = 0

    def _controllerType(self, direction: int):
        if self._cursor_position == len(self._save_name):
            if self._cursor_position < (self._MENU_CHAR_WIDTH - 1):
                char_pos = max(direction - 1, -1)
                self._save_name += self._CONTROLLER_CHARS[char_pos]
                self._resetCursorBlink()
        else:
            char_pos = self._CONTROLLER_CHARS.find(
                self._save_name[self._cursor_position].lower()
            )
            new_char_pos = (char_pos + direction) % len(self._CONTROLLER_CHARS)
            self._save_name = self._save_name[:self._cursor_position] \
                + self._CONTROLLER_CHARS[new_char_pos] \
                + self._save_name[self._cursor_position + 1:]

    def _backspace(self):
        self._save_name = self._save_name[:self._cursor_position - 1] + self._save_name[self._cursor_position:]
        self._cursor_position -= 1
        self._resetCursorBlink()

    def _inputEvent(self, event):
        match self._getAction(event):
            case MenuAction.QUIT:
                self.next_mode = ModeGameMenuTop(self._previous_mode, self._background)
            case MenuAction.REJECT:
                if self._confirm_overwrite:
                    self._confirm_overwrite = False
                    self._save_success = None
                elif event.type == pygame.JOYBUTTONDOWN and self._cursor_position > 0:
                    self._backspace()
                else:
                    self.next_mode = ModeGameMenuTop(self._previous_mode, self._background)
            case MenuAction.CONFIRM:
                if self._save_success:
                    self.next_mode = ModeGameMenuTop(self._previous_mode, self._background)
                elif isinstance(self._previous_mode, Saveable):
                    if not self._save_name:
                        self._save_name = utility.getDateTimeFileName()
                    if Save.willOverwrite(self._save_name) and not self._confirm_overwrite:
                        self._confirm_overwrite = True
                    else:
                        new_save = Save.getFromMode(self._save_name, self._previous_mode)
                        self._save_success = new_save.save()
            case MenuAction.LEFT:
                self._cursor_position -= 1
                self._resetCursorBlink()
            case MenuAction.RIGHT:
                self._cursor_position += 1
                self._resetCursorBlink()
            case MenuAction.UP:
                if event.type == pygame.JOYHATMOTION:
                    self._controllerType(-1)
            case MenuAction.DOWN:
                if event.type == pygame.JOYHATMOTION:
                    self._controllerType(1)
        if event.type == pygame.KEYDOWN:
            match event.key:
                case pygame.K_UP | pygame.K_HOME | pygame.K_PAGEUP:
                    self._cursor_position = 0
                    self._resetCursorBlink()
                case pygame.K_DOWN | pygame.K_END | pygame.K_PAGEDOWN:
                    self._cursor_position = len(self._save_name)
                    self._resetCursorBlink()
                case pygame.K_DELETE:
                    self._save_name = self._save_name[:self._cursor_position] \
                        + self._save_name[self._cursor_position + 1:]
                    self._resetCursorBlink()
                case pygame.K_BACKSPACE:
                    self._backspace()
                case _:
                    if (
                        len(self._save_name) < (self._MENU_CHAR_WIDTH - 1)
                        and (
                            # numbers
                            ('0' <= event.unicode <= '9')
                            # or letters
                            or (96 < event.key < 123)
                        )
                    ):
                        self._save_name = self._save_name[:self._cursor_position] \
                            + event.unicode \
                            + self._save_name[self._cursor_position:]
                        self._cursor_position += 1
                        self._resetCursorBlink()
        self._cursor_position = utility.clamp(self._cursor_position, 0, len(self._save_name))

    def _update(self, dt):
        self._cursor_timer += dt
        if self._cursor_timer >= self._CURSOR_TIME:
            self._cursor_switch = not self._cursor_switch
            self._cursor_timer -= self._CURSOR_TIME

    def _drawPreSprites(self, screen):
        disp_text = self._SHARED_DISP_TEXT
        draw_cursor = False
        if not isinstance(self._previous_mode, Saveable):
            disp_text += "\nYou can't save now."
        elif self._save_success:
            disp_text += "\nSaved successfully.\nPress ENTER to continue."
        else:
            disp_text += "ENTER) Save\nType a save name (or leave blank for default):\n>"
            if self._save_name:
                disp_text += self._save_name
            if self._confirm_overwrite and self._save_success is None:
                disp_text += "\nThis will overwrite an existing save." \
                    + "\nPress ENTER again to confirm, or ESC to go back."
            elif self._save_success is False:
                disp_text += "\nSave failed.\nPress ENTER to try again, or ESC to go back."
            else:
                draw_cursor = True
        self._drawTextAlways(disp_text)
        if self._cursor_switch and draw_cursor:
            cursor_x = getDefaultFontWrap().font.size(">" + self._save_name[:self._cursor_position])[0]
            cursor_y = self._menu_surface.get_height() - getDefaultFontWrap().line_height
            self._menu_surface.fill(
                self._TEXT_COLOR,
                (
                    (cursor_x, cursor_y),
                    (1, getDefaultFontWrap().line_height)
                )
            )
        screen.blit(self._menu_surface, (0, 0))


class ModeGameMenuLoad(ModeGameMenuList):
    STATE_DEFAULT = 0
    STATE_LOADED_SAVE = 1
    STATE_DELETED_SAVE = 2
    STATE_SELECTED_SAVE = 3
    OPTION_LOAD = 0
    OPTION_DELETE = 1

    __slots__ = (
        '_saves',
        '_state',
        '_selected_save_option',
    )

    def __init__(self, previous_mode, old_screen):
        super().__init__(previous_mode, old_screen)
        self._saves = Save.getAllFromFiles()
        self._state = self.STATE_DEFAULT
        self._selected_save_option = self.OPTION_LOAD

    def _inputEvent(self, event):
        action = self._getAction(event)
        if action == MenuAction.QUIT:
            self.next_mode = ModeGameMenuTop(self._previous_mode, self._background)
            return
        match self._state:
            case self.STATE_DEFAULT:
                match action:
                    case MenuAction.REJECT:
                        self.next_mode = ModeGameMenuTop(self._previous_mode, self._background)
                    case MenuAction.UP | MenuAction.LEFT:
                        self._index -= 1
                    case MenuAction.DOWN | MenuAction.RIGHT:
                        self._index += 1
                    case MenuAction.CONFIRM:
                        if len(self._saves) > 0:
                            self._state = self.STATE_SELECTED_SAVE
                            self._selected_save_option = self.OPTION_LOAD
                self._index = utility.clamp(self._index, 0, len(self._saves) - 1)
            case self.STATE_LOADED_SAVE:
                if action in (MenuAction.CONFIRM, MenuAction.REJECT):
                    self.next_mode = ModeGameMenuTop(self._previous_mode, self._background)
            case self.STATE_DELETED_SAVE:
                if action in (MenuAction.CONFIRM, MenuAction.REJECT):
                    self._state = self.STATE_DEFAULT
            case self.STATE_SELECTED_SAVE:
                match action:
                    case MenuAction.UP | MenuAction.LEFT:
                        self._selected_save_option -= 1
                        self._selected_save_option %= 2
                    case MenuAction.DOWN | MenuAction.RIGHT:
                        self._selected_save_option += 1
                        self._selected_save_option %= 2
                    case MenuAction.CONFIRM:
                        if self._selected_save_option == self.OPTION_LOAD:
                            self._stopMixer()
                            self._previous_mode = self._saves[self._index].load()
                            pygame.mixer.music.pause()
                            pygame.mixer.pause()
                            self._background = self._getOldScreen()
                            self._state = self.STATE_LOADED_SAVE
                        elif self._selected_save_option == self.OPTION_DELETE:
                            self._saves[self._index].delete()
                            del self._saves[self._index]
                            self._index = utility.clamp(self._index, 0, len(self._saves) - 1)
                            self._state = self.STATE_DELETED_SAVE
                    case MenuAction.REJECT:
                        self._state = self.STATE_DEFAULT

    def _getOptionsLength(self):
        return len(self._saves)

    def _getOptionName(self, index):
        return self._saves[index].save_name

    def _getOptionStatus(self, option: int):
        return self._getSelectedChar(self._selected_save_option == option)

    def _drawPreSprites(self, screen):
        disp_text = self._SHARED_DISP_TEXT
        match self._state:
            case self.STATE_DEFAULT:
                if len(self._saves) == 0:
                    disp_text += "\nThere are no saves to select from."
                else:
                    disp_text += "ARROW KEYS + ENTER) Select a save:"
                    disp_text += self._getOptionsText()
            case self.STATE_LOADED_SAVE:
                disp_text += "\nLoaded successfully.\nPress ENTER to continue."
            case self.STATE_DELETED_SAVE:
                disp_text += "\nDeleted successfully.\nPress ENTER to continue."
            case self.STATE_SELECTED_SAVE:
                disp_text += "ARROW KEYS + ENTER) Select a save:"
                disp_text += self._getOptionsText()
                disp_text += f"\n{self._getOptionStatus(self.OPTION_LOAD)}Load" \
                    + f"\n{self._getOptionStatus(self.OPTION_DELETE)}Delete"
        self._drawText(disp_text)
        screen.blit(self._menu_surface, (0, 0))


class ModeGameMenuOptions(ModeGameMenu):
    __slots__ = (
        '_pressed_return',
    )

    def __init__(self, previous_mode, old_screen, pressed_return: bool):
        super().__init__(previous_mode, old_screen)
        self._pressed_return = pressed_return

    def _getAction(self, event: pygame.event.Event):
        if event.type == pygame.KEYDOWN and event.key != pygame.K_ESCAPE:
            return MenuAction.NOTHING
        elif event.type == pygame.KEYUP:
            match event.key:
                case pygame.K_LEFT:
                    return MenuAction.LEFT
                case pygame.K_RIGHT:
                    return MenuAction.RIGHT
                case pygame.K_UP:
                    return MenuAction.UP
                case pygame.K_DOWN:
                    return MenuAction.DOWN
                case pygame.K_RETURN:
                    if self._pressed_return:
                        self._pressed_return = False
                    else:
                        return MenuAction.CONFIRM
        return super()._getAction(event)

    def _inputEvent(self, event):
        match self._getAction(event):
            case MenuAction.QUIT | MenuAction.REJECT:
                self.next_mode = ModeGameMenuTop(self._previous_mode, self._background)
            case MenuAction.DOWN | MenuAction.LEFT:
                display.changeScale(-1)
            case MenuAction.UP | MenuAction.RIGHT:
                display.changeScale(1)
            case MenuAction.CONFIRM:
                display.toggleFullscreen()
        if event.type == pygame.KEYDOWN and '1' <= event.unicode <= '9':
            target_scale = int(event.unicode)
            display.setScale(target_scale)

    def _drawPreSprites(self, screen):
        disp_text = self._SHARED_DISP_TEXT
        disp_text += f"ARROW KEYS) Upscaling: {display.upscale}" \
            + f"\nENTER) Fullscreen: {self.getTickBox(display.is_fullscreen)}"
        self._drawText(disp_text)
        screen.blit(self._menu_surface, (0, 0))

    @staticmethod
    def getTickBox(value: bool):
        inside = "*" if value else "_"
        return f"[{inside}]"


class ModeGameMenuControls(ModeGameMenuList):
    STATE_CHOOSE_PLAYER = 0
    STATE_CHOOSE_EVENT = 1
    STATE_CHOOSE_INPUT = 2

    __slots__ = (
        '_state',
        '_selected_player',
        '_selected_input',
        '_selection_timer',
    )

    def __init__(self, previous_mode, old_screen):
        super().__init__(previous_mode, old_screen)
        self._state = self.STATE_CHOOSE_PLAYER if self._mustSelectPlayer() else self.STATE_CHOOSE_EVENT
        self._selected_player = 0
        self._selection_timer = 5000

    @staticmethod
    def _mustSelectPlayer() -> bool:
        return input.max_players != 1

    def _getOptionsLength(self):
        return input.num_inputs

    def _getOptionName(self, index):
        return input.getEventWithControls(self._selected_player, index)

    def _inputEvent(self, event):
        action = self._getAction(event)
        if action == MenuAction.QUIT:
            self.next_mode = ModeGameMenuTop(self._previous_mode, self._background)
            return
        match self._state:
            case self.STATE_CHOOSE_PLAYER:
                match action:
                    case MenuAction.UP | MenuAction.LEFT:
                        self._selected_player -= 1
                    case MenuAction.DOWN | MenuAction.RIGHT:
                        self._selected_player += 1
                    case MenuAction.CONFIRM:
                        self._state = self.STATE_CHOOSE_EVENT
                        self._index = 0
                    case MenuAction.REJECT:
                        self.next_mode = ModeGameMenuTop(self._previous_mode, self._background)
                self._selected_player = utility.clamp(self._selected_player, 0, input.max_players - 1)
            case self.STATE_CHOOSE_EVENT:
                match action:
                    case MenuAction.UP | MenuAction.LEFT:
                        self._index -= 1
                    case MenuAction.DOWN | MenuAction.RIGHT:
                        self._index += 1
                    case MenuAction.CONFIRM:
                        self._state = self.STATE_CHOOSE_INPUT
                        self._selection_timer = 5000
                    case MenuAction.REJECT:
                        if self._mustSelectPlayer():
                            self._state = self.STATE_CHOOSE_PLAYER
                        else:
                            self.next_mode = ModeGameMenuTop(self._previous_mode, self._background)
                self._index = utility.clamp(self._index, 0, input.num_inputs - 1)
            case self.STATE_CHOOSE_INPUT:
                if input.setInputMapping(self._selected_player, self._index, event):
                    self._state = self.STATE_CHOOSE_EVENT

    def _update(self, dt):
        if self._state == self.STATE_CHOOSE_INPUT:
            self._selection_timer -= dt
            if self._selection_timer <= 0:
                self._state = self.STATE_CHOOSE_EVENT

    def _drawPreSprites(self, screen):
        disp_text = self._SHARED_DISP_TEXT
        if self._state != self.STATE_CHOOSE_INPUT:
            disp_text += "ARROW KEYS + ENTER)\n"
        if self._mustSelectPlayer():
            disp_text += f"Player: {self._selected_player + 1}\n"
        if self._state == self.STATE_CHOOSE_EVENT:
            disp_text += "Action:"
            disp_text += self._getOptionsText()
        if self._state == self.STATE_CHOOSE_INPUT:
            disp_text += f"Action: {input.getEventName(self._index)}"
            disp_text += "\n\n____press a button to select"
            disp_text += f"\n____(wait {(self._selection_timer // 1000) + 1} seconds to exit)"
        self._drawText(disp_text)
        screen.blit(self._menu_surface, (0, 0))



from collections.abc import Collection


class ControllerStateChange(object):
    __slots__ = (
        'player_id',
        'event_type',
        'new_value',
    )

    def __init__(self, player_id: int, event_type: int, new_value: float | int):
        self.player_id = player_id
        self.event_type = event_type
        self.new_value = new_value


class InputFrame(object):
    __slots__ = (
        '_controller_states',
        '_controller_states_prev',
        '_controller_state_changes',
    )

    def __init__(
        self,
        controller_states: list[list[float | int]],
        controller_states_prev: list[list[float | int]],
        controller_state_changes: list[ControllerStateChange]
    ):
        self._controller_states = controller_states
        self._controller_states_prev = controller_states_prev
        self._controller_state_changes = controller_state_changes

    def getInputState(self, player_id: int, event_type: int):
        return self._controller_states[player_id][event_type]

    def wasPlayerInputPressed(self, player_id: int, event_type: int):
        for state_change in self._controller_state_changes:
            if state_change.new_value == 1 \
                    and state_change.event_type == event_type \
                    and state_change.player_id == player_id:
                return True
        return False

    def wasInputPressed(self, event_type: int):
        for state_change in self._controller_state_changes:
            if state_change.new_value == 1 \
                    and state_change.event_type == event_type:
                return True
        return False

    def wasAnyPlayerInputPressed(self, player_id: int, event_types: Collection[int]):
        for state_change in self._controller_state_changes:
            if state_change.new_value == 1 \
                    and state_change.event_type in event_types \
                    and state_change.player_id == player_id:
                return True
        return False

    def wasAnyInputPressed(self, event_types: Collection[int]):
        for state_change in self._controller_state_changes:
            if state_change.new_value == 1 \
                    and state_change.event_type in event_types:
                return True
        return False



import pygame


class FontWrap(object):
    __slots__ = (
        'font',
        'line_height',
        '_antialias',
    )

    def __init__(self, font: pygame.font.Font, line_height: int, antialias: bool):
        self.font = font
        self.line_height = line_height
        self._antialias = antialias

    def renderTo(self, surf: pygame.Surface, dest, text: str, color, background=None):
        surf.blit(self.font.render(text, self._antialias, color, background), dest)

    def renderToCentered(self, surf: pygame.Surface, dest, text: str, color, background=None):
        text_size = self.font.size(text)
        surf.blit(
            self.font.render(text, self._antialias, color, background),
            (dest[0] - text_size[0] // 2, dest[1] - text_size[1] // 2)
        )

    def _calculateLinesForWords(self, width: int, words: list[str]):
        lines = [words[0].replace('_', ' ')]
        for word in words[1:]:
            new_word = word.replace('_', ' ')
            if self.font.size(lines[-1] + " " + new_word)[0] > width:
                lines.append(new_word)
            else:
                lines[-1] += " " + new_word
        return lines

    def renderToInside(self, surf: pygame.Surface, dest, width: int, text: str, color, background=None):
        # probably more efficient to do once?
        part_dest = [dest[0], dest[1]]
        for words in [line.split() for line in text.splitlines()]:
            if not words:
                words = [""]
            lines = self._calculateLinesForWords(width, words)
            for line in lines:
                surf.blit(
                    self.font.render(line, self._antialias, color, background),
                    part_dest
                )
                part_dest[1] += self.line_height

    def _renderWordsInside(self, width: int, words: list[str], color, background):
        """Returns a surface of the width with the words drawn on it.
        If any word is too long to fit, it will be in its own line, and truncated.
        """
        lines = self._calculateLinesForWords(width, words)
        result = pygame.Surface((width, self.line_height * len(lines))).convert()
        result.fill(background)
        for i, line in enumerate(lines):
            drawn_line = self.font.render(line, self._antialias, color, background).convert()
            result.blit(drawn_line, (0, i * self.line_height))
        return result

    def renderInside(self, width: int, text: str, color, background):
        # probably more efficient if keeping resultant surface and using that to draw over and over?
        height = 0
        imgs = []
        for words in [line.split() for line in text.splitlines()]:
            if not words:
                words = [""]
            imgs.append(self._renderWordsInside(width, words, color, background))
            height += imgs[-1].get_height()
        result = pygame.Surface((width, height)).convert()
        dest = [0, 0]
        for img in imgs:
            result.blit(img, dest)
            dest[1] += img.get_height()
        return result


_default: FontWrap | None = None


def init(font: pygame.font.Font, line_height: int, antialias: bool):
    global _default
    if _default:
        raise RuntimeError("error: _default is already set")
    _default = FontWrap(font, line_height, antialias)


def getDefaultFontWrap():
    return _default



import functools

import pygame


@functools.cache
def image(filename, alpha_or_colorkey=False):
    """Loads images, converts, and sets colorkey as needed.
    The results are cached, so don't blit to them."""
    result = pygame.image.load(filename)
    if alpha_or_colorkey is True:
        result = result.convert_alpha()
    else:
        result = result.convert()
        if alpha_or_colorkey is not False:
            result.set_colorkey(alpha_or_colorkey)
    return result


@functools.cache
def sound(filename):
    """Loads sounds.
    The results are cached, so don't set volume on them."""
    return pygame.mixer.Sound(filename)



import os
import copy
import enum
from collections.abc import Iterable

import pygame

from .inputframe import ControllerStateChange, InputFrame


class InputType(enum.Enum):
    KEYBOARD = enum.auto()
    MOUSE = enum.auto()
    CON_BUTTON = enum.auto()
    CON_HAT = enum.auto()
    CON_AXIS = enum.auto()


class InputDefault(object):
    __slots__ = (
        'player_id',
        'event_type',
        'input_type',
        'input_id',
        'controller_id',
    )

    def __init__(
        self,
        player_id: int,
        event_type: int,
        input_type: InputType,
        input_id: int,
        controller_id: int = 0
    ):
        self.player_id = player_id
        self.event_type = event_type
        self.input_type = input_type
        self.input_id = input_id
        self.controller_id = controller_id

    def getMapKey(self):
        return _getMapKey(
            self.input_type,
            self.input_id,
            self.controller_id
        )

    def getMapValue(self):
        return (
            self.player_id,
            self.event_type,
        )


_HAT_BUTTON_NAMES = (
    "L",
    "R",
    "U",
    "D",
)
TYPE_NONE = -1
TYPE_PAUSE = 0
TYPE_SCREENSHOT = 1
_ENGINE_INPUT_NAMES = (
    "Pause",
    "Screenshot",
)
_ENGINE_INPUT_DEFAULTS = (
    InputDefault(0, TYPE_PAUSE, InputType.KEYBOARD, pygame.K_ESCAPE),
    InputDefault(0, TYPE_SCREENSHOT, InputType.KEYBOARD, pygame.K_F12),
)
_CONTROLLER_PAUSE_BUTTON = 7
EVENT_TYPE_START_POS = len(_ENGINE_INPUT_NAMES)
_input_file: str | None = None
max_players: int
_event_names: tuple[str]
num_inputs: int
_input_defaults: tuple[InputDefault]
_input_mapping: dict[tuple[InputType, int, int], tuple[int, int]]
_controller_states: list[list[float | int]]
_controller_states_prev: list[list[float | int]]
_controller_state_changes: list[ControllerStateChange]


def init(input_file: str, max_players_in: int, event_names: tuple[str], input_defaults: tuple[InputDefault]):
    global _input_file
    global max_players
    global _event_names
    global num_inputs
    global _input_defaults
    if _input_file:
        raise RuntimeError("error: _input_file is already set")
    if max_players_in < 1:
        raise ValueError("error: max_players_in must be at least 1")
    _input_file = input_file
    max_players = max_players_in
    _event_names = _ENGINE_INPUT_NAMES + event_names
    num_inputs = len(_event_names)
    _input_defaults = input_defaults
    if os.path.exists(_input_file):
        _load()
    else:
        resetDefaultMapping()
    startNewMode()


def startNewMode():
    global _controller_states
    global _controller_states_prev
    global _controller_state_changes
    _controller_states = [[0] * num_inputs for _ in range(max_players)]
    _controller_states_prev = [[0] * num_inputs for _ in range(max_players)]
    _controller_state_changes = []


def _setInputMapping(input_default: InputDefault):
    _input_mapping[input_default.getMapKey()] = input_default.getMapValue()


def resetDefaultMapping():
    global _input_mapping
    _input_mapping = dict()
    controller_pause = tuple(
        InputDefault(i, TYPE_PAUSE, InputType.CON_BUTTON, _CONTROLLER_PAUSE_BUTTON, i)
        for i in range(max_players)
    )
    for input_default in _ENGINE_INPUT_DEFAULTS + controller_pause + _input_defaults:
        _setInputMapping(input_default)


def _getHatValues(event: pygame.event.Event) -> tuple[int, int, int, int]:
    return (
        1 if event.value[0] == -1 else 0,
        1 if event.value[0] == 1 else 0,
        1 if event.value[1] == 1 else 0,
        1 if event.value[1] == -1 else 0,
    )


def setInputMapping(player_id: int, event_type: int, event: pygame.event.Event) -> bool:
    match event.type:
        case pygame.KEYDOWN:
            input_default = InputDefault(
                player_id,
                event_type,
                InputType.KEYBOARD,
                event.key
            )
        case pygame.MOUSEBUTTONDOWN:
            input_default = InputDefault(
                player_id,
                event_type,
                InputType.MOUSE,
                event.button
            )
        case pygame.JOYBUTTONDOWN:
            input_default = InputDefault(
                player_id,
                event_type,
                InputType.CON_BUTTON,
                event.button,
                event.instance_id
            )
        case pygame.JOYHATMOTION:
            input_id = None
            for i, event_value in enumerate(_getHatValues(event)):
                if event_value == 1:
                    input_id = event.hat * 4 + i
            input_default = InputDefault(
                player_id,
                event_type,
                InputType.CON_HAT,
                input_id,
                event.instance_id
            )
        case _:
            return False
    _setInputMapping(input_default)
    return True


def _getInputIdDisplay(input_type: InputType, input_id: int) -> str:
    match input_type:
        case InputType.KEYBOARD:
            return pygame.key.name(input_id)
        case InputType.CON_HAT:
            return f'{input_id // 4}{_HAT_BUTTON_NAMES[input_id % 4]}'
        case _:
            return str(input_id)


def _getDisplayInput(input_type: InputType, input_id: int, controller_id: int) -> str:
    match input_type:
        case InputType.KEYBOARD:
            result = f'KEY-'
        case InputType.MOUSE:
            result = f'MOUSE-'
        case InputType.CON_BUTTON:
            result = f'CON{controller_id}-BT-'
        case InputType.CON_HAT:
            result = f'CON{controller_id}-HT-'
        case InputType.CON_AXIS:
            result = f'CON{controller_id}-AX-'
        case _:
            raise ValueError("error: input_type must be a supported InputType")
    result += _getInputIdDisplay(input_type, input_id)
    return result


def getEventWithControls(player_id: int, event_type: int) -> str:
    inputs = [
        _getDisplayInput(*key)
        for key, value in _input_mapping.items()
        if value == (player_id, event_type)
    ]
    return f"{_event_names[event_type]}: {','.join(inputs)}"


def getEventName(event_type: int) -> str:
    return _event_names[event_type]


_PLAYER_SEP = '|'
_EVENT_SEP = ':'
_INPUT_SEP = '&'
_PART_SEP = '+'


def _load():
    global _input_mapping
    _input_mapping = dict()
    with open(_input_file, 'r') as file:
        for line in file:
            line_parts = line.strip().split(_PLAYER_SEP)
            player_id = int(line_parts[0].strip())
            line_parts = line_parts[1].strip().split(_EVENT_SEP)
            event_name = line_parts[0].strip()
            event_type = _event_names.index(event_name)
            input_sections = line_parts[1].strip().split(_INPUT_SEP)
            for input_section in input_sections:
                if not input_section:
                    continue
                input_parts = input_section.strip().split(_PART_SEP)
                input_type_name = input_parts[0]
                input_type = InputType[input_type_name]
                input_id_or_name = input_parts[1]
                input_id: int
                match input_type:
                    case InputType.KEYBOARD:
                        input_id = pygame.key.key_code(input_id_or_name)
                    case InputType.CON_HAT:
                        hat_number = int(input_id_or_name[:-1])
                        hat_direction = input_id_or_name[-1]
                        input_id = hat_number * 4 + _HAT_BUTTON_NAMES.index(hat_direction)
                    case _:
                        input_id = int(input_id_or_name)
                controller_id = 0
                if len(input_parts) == 3:
                    controller_id = int(input_parts[2])
                _input_mapping[_getMapKey(input_type, input_id, controller_id)] = (player_id, event_type)


def _getSaveInput(input_type: InputType, input_id: int, controller_id: int):
    result = f'{input_type.name}{_PART_SEP}'
    result += _getInputIdDisplay(input_type, input_id)
    if input_type in (InputType.CON_BUTTON, InputType.CON_HAT, InputType.CON_AXIS):
        result += f'{_PART_SEP}{controller_id}'
    return result


def save():
    mapping_to_write: dict[tuple[int, int], list[tuple[InputType, int, int]]] = dict()
    for key, value in _input_mapping.items():
        if value not in mapping_to_write:
            mapping_to_write[value] = []
        mapping_to_write[value].append(key)
    with open(_input_file, 'w') as file:
        for write_key, write_value in mapping_to_write.items():
            inputs = _INPUT_SEP.join(
                [
                    _getSaveInput(item[0], item[1], item[2])
                    for item
                    in write_value
                ]
            )
            line = ('{}' + _PLAYER_SEP + '{}' + _EVENT_SEP + '{}').format(
                write_key[0],
                _event_names[write_key[1]],
                inputs
            )
            print(line, file=file)


def _takeEvent(event: pygame.event.Event):
    match event.type:
        case pygame.KEYUP | pygame.KEYDOWN:
            player_id, event_type = _mapEvent(InputType.KEYBOARD, event.key)
            return _logEvent(player_id, event_type, 1 if event.type == pygame.KEYDOWN else 0)
        case pygame.MOUSEBUTTONUP | pygame.MOUSEBUTTONDOWN:
            player_id, event_type = _mapEvent(InputType.MOUSE, event.button)
            return _logEvent(player_id, event_type, 1 if event.type == pygame.MOUSEBUTTONDOWN else 0)
        case pygame.JOYBUTTONUP | pygame.JOYBUTTONDOWN:
            player_id, event_type = _mapEvent(InputType.CON_BUTTON, event.button, event.instance_id)
            return _logEvent(player_id, event_type, 1 if event.type == pygame.JOYBUTTONDOWN else 0)
        case pygame.JOYHATMOTION:
            for i, event_value in enumerate(_getHatValues(event)):
                player_id, event_type = _mapEvent(InputType.CON_HAT, event.hat * 4 + i, event.instance_id)
                _logEvent(player_id, event_type, event_value)
        case pygame.JOYAXISMOTION:
            axis_value_back_forth = (
                event.value * -1 if event.value < 0 else 0,
                event.value if event.value > 0 else 0,
            )
            for i, event_value in enumerate(axis_value_back_forth):
                player_id, event_type = _mapEvent(InputType.CON_AXIS, event.axis * 2 + i, event.instance_id)
                _logEvent(player_id, event_type, event_value)
    return True


def takeEvents(events: Iterable[pygame.event.Event]):
    global _controller_states_prev
    global _controller_state_changes
    _controller_states_prev = copy.deepcopy(_controller_states)
    _controller_state_changes = []
    events = filter(_takeEvent, events)
    return list(events)


def _getMapKey(input_type: InputType, input_id: int, controller_id: int):
    return (
        input_type,
        input_id,
        controller_id,
    )


def _mapEvent(input_type: InputType, input_id: int, controller_id: int = 0):
    return _input_mapping.get(
        _getMapKey(input_type, input_id, controller_id),
        (0, TYPE_NONE,)
    )


def _logEvent(player_id: int, event_type: int, event_value: float | int):
    if event_type != TYPE_NONE and _controller_states[player_id][event_type] != event_value:
        _controller_state_changes.append(
            ControllerStateChange(player_id, event_type, event_value)
        )
        _controller_states[player_id][event_type] = event_value
    return event_type != TYPE_SCREENSHOT


def getInputFrame():
    return InputFrame(_controller_states, _controller_states_prev, _controller_state_changes)



import unittest
import os

import pygame

import jovialengine.input as input


class TestInput(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        input.init(
            os.path.join('.', 'input.cfg'),
            1,
            (
                "Left",
                "Right",
                "Up",
                "Down",
                "Confirm",
                "Cancel",
            ),
            (
                input.InputDefault(0, input.EVENT_TYPE_START_POS, input.InputType.KEYBOARD, pygame.K_a),
            )
        )

    def test_init(self):
        # Assert
        self.assertEqual(len(input._controller_states), 1)
        self.assertEqual(len(input._controller_states[0]), 8)


if __name__ == '__main__':
    unittest.main()



import random
import math
from datetime import datetime


def clamp(number, minimum, maximum):
    return max(minimum, min(maximum, number))


def getIntMovement(tracking, vel, dt):
    tracking += vel * dt
    tracking_int = int(tracking)
    tracking -= tracking_int
    return tracking, tracking_int


def reduceNumber(number, divisor):
    result = number // divisor
    mod = number % divisor
    if random.random() < (mod / divisor):
        result += 1
    return result


def sinCurve(number: float | int):
    return math.sin(number * math.pi / 2)


def cosCurve(number: float | int):
    if number == 1:
        return 0
    return math.cos(number * math.pi / 2)


def binary(start, end, mix):
    if mix < 1.0:
        return start
    return end


def lerp(start, end, mix):
    return start + (end - start)*mix


def incSpeedLerp(start, end, mix):
    return lerp(start, end, mix**2)


def decSpeedLerp(start, end, mix):
    return lerp(start, end, 1 - (1 - mix)**2)


def incDecSpeedLerp(start, end, mix):
    midpoint = lerp(start, end, 0.5)
    if mix < 0.5:
        return incSpeedLerp(start, midpoint, mix*2)
    else:
        return decSpeedLerp(midpoint, end, (mix - 0.5)*2)


def decIncSpeedLerp(start, end, mix):
    midpoint = lerp(start, end, 0.5)
    if mix < 0.5:
        return decSpeedLerp(start, midpoint, mix*2)
    else:
        return incSpeedLerp(midpoint, end, (mix - 0.5)*2)


def getDateTimeFileName():
    return datetime.utcnow().isoformat().replace(':', '')



import configparser


MAX_FRAMERATE = 'MaxFramerate'
FULLSCREEN = 'Fullscreen'
SCREEN_SCALE = 'ScreenScale'
_DEFAULTS = {
    MAX_FRAMERATE: 1000,
    SCREEN_SCALE: 0,
    FULLSCREEN: False,
}
_SECTION = 'Game'
_config = configparser.ConfigParser(_DEFAULTS, default_section=_SECTION)
_config_file: str | None = None


def init(config_file: str):
    global _config_file
    if _config_file:
        raise RuntimeError("error: _config_file is already set")
    _config_file = config_file
    _config.read(_config_file)
    for section in _config.sections():
        _config.remove_section(section)


def get(key: str):
    match _DEFAULTS.get(key):
        case bool():
            return _config.getboolean(_SECTION, key)
        case int():
            return _config.getint(_SECTION, key)
        case _:
            return _config.get(_SECTION, key)


def update(key: str, value):
    _config.set(_SECTION, key, str(value))


def save():
    with open(_config_file, 'w') as file:
        _config.write(file, space_around_delimiters=False)



import abc
from typing import final
from collections.abc import Iterable

import pygame

from .inputframe import InputFrame


class ModeBase(abc.ABC):
    """This is an abstract object for game modes.
    """
    __slots__ = (
        '_space',
        '_background',
        '_all_sprites',
        '_camera',
        '_input_frame',
        'next_mode',
    )

    @abc.abstractmethod
    def __init__(self):
        """Implementation must contain a call to _init passing in the space size before all else."""
        raise NotImplementedError(
            type(self).__name__ + ".__init__(self)"
        )

    def _init(self, space_size: tuple[int, int]):
        self._space = pygame.Surface(space_size).convert()
        self._background = pygame.Surface(space_size).convert()
        self._background.fill((255, 255, 255))
        self._all_sprites = pygame.sprite.LayeredDirty()
        self._camera = pygame.rect.Rect((0, 0), space_size)
        self._input_frame: InputFrame | None = None
        """All game modes must set the next mode when they are done.
        Don't create another mode unless you are immediately assigning it to self.next_mode
        """
        self.next_mode: ModeBase | None = None

    def cleanup(self):
        self._all_sprites.empty()

    def _inputEvent(self, event: pygame.event.Event):
        """Handle any input that requires looking at pygame events directly, like typing."""
        pass

    def _inputFrame(self, input_frame: InputFrame):
        """Handle all other input."""
        pass

    @final
    def input(self, events: Iterable[pygame.event.Event], input_frame: InputFrame):
        """All game modes can take in input."""
        for event in events:
            self._inputEvent(event)
        self._inputFrame(input_frame)
        self._input_frame = input_frame

    def _update(self, dt: int):
        pass

    @final
    def update(self, dt: int):
        """All game modes can update."""
        self._update(dt)
        self._all_sprites.update(dt)

    def _updatePreDraw(self, screen: pygame.surface.Surface):
        pass

    def _drawPreSprites(self, screen: pygame.surface.Surface):
        pass

    def _drawPostSprites(self, screen: pygame.surface.Surface):
        pass

    def _drawPostCamera(self, screen: pygame.surface.Surface):
        pass

    @final
    def draw(self, screen: pygame.surface.Surface):
        """All game modes can draw to the screen"""
        self._updatePreDraw(screen)
        self._space.blit(self._background, (0, 0))
        self._drawPreSprites(self._space)
        self._all_sprites.draw(self._space)
        self._drawPostSprites(self._space)
        screen.blit(self._space, (0, 0), self._camera)
        self._drawPostCamera(screen)

    @staticmethod
    def _stopMixer():
        pygame.mixer.music.stop()
        pygame.mixer.music.unload()
        pygame.mixer.stop()



import unittest

import jovialengine.utility as utility


class TestUtility(unittest.TestCase):
    def test_clamp(self):
        self.assertEqual(
            utility.clamp(123, 5, 10),
            10
        )
        self.assertEqual(
            utility.clamp(-123, 5, 10),
            5
        )
        self.assertEqual(
            utility.clamp(12, 3, 33),
            12
        )

    def test_getIntMovement(self):
        tracking = 0.6
        tracking, tracking_int = utility.getIntMovement(tracking, 3.75, 10)
        self.assertEqual(tracking_int, 38)
        self.assertTrue(0.0999999999 < tracking < 0.1000000001)

    def test_binary(self):
        self.assertEqual(
            utility.binary(1.0, 2.0, 0.0),
            1
        )
        self.assertEqual(
            utility.binary(1.0, 2.0, 0.99999),
            1
        )
        self.assertEqual(
            utility.binary(1.0, 2.0, 1.0),
            2
        )
        self.assertEqual(
            utility.binary(1.0, 2.0, 1),
            2
        )

    def test_lerp(self):
        self.assertEqual(
            utility.lerp(1.0, 2.0, 0.0),
            1
        )
        self.assertEqual(
            utility.lerp(1.0, 2.0, 0.25),
            1.25
        )
        self.assertEqual(
            utility.lerp(1.0, 2.0, 0.75),
            1.75
        )
        self.assertEqual(
            utility.lerp(1.0, 2.0, 1.0),
            2
        )

    def test_incSpeedLerp(self):
        self.assertEqual(
            utility.incSpeedLerp(1.0, 2.0, 0.0),
            1
        )
        self.assertEqual(
            utility.incSpeedLerp(1.0, 2.0, 0.25),
            1.0625
        )
        self.assertEqual(
            utility.incSpeedLerp(1.0, 2.0, 0.75),
            1.5625
        )
        self.assertEqual(
            utility.incSpeedLerp(1.0, 2.0, 1.0),
            2
        )

    def test_decSpeedLerp(self):
        self.assertEqual(
            utility.decSpeedLerp(1.0, 2.0, 0.0),
            1
        )
        self.assertEqual(
            utility.decSpeedLerp(1.0, 2.0, 0.25),
            1.4375
        )
        self.assertEqual(
            utility.decSpeedLerp(1.0, 2.0, 0.75),
            1.9375
        )
        self.assertEqual(
            utility.decSpeedLerp(1.0, 2.0, 1.0),
            2
        )

    def test_incDecSpeedLerp(self):
        self.assertEqual(
            utility.incDecSpeedLerp(1.0, 2.0, 0.0),
            1
        )
        self.assertEqual(
            utility.incDecSpeedLerp(1.0, 2.0, 0.25),
            1.125
        )
        self.assertEqual(
            utility.incDecSpeedLerp(1.0, 2.0, 0.75),
            1.875
        )
        self.assertEqual(
            utility.incDecSpeedLerp(1.0, 2.0, 1.0),
            2
        )

    def test_decIncSpeedLerp(self):
        self.assertEqual(
            utility.decIncSpeedLerp(1.0, 2.0, 0.0),
            1
        )
        self.assertEqual(
            utility.decIncSpeedLerp(1.0, 2.0, 0.25),
            1.375
        )
        self.assertEqual(
            utility.decIncSpeedLerp(1.0, 2.0, 0.75),
            1.625
        )
        self.assertEqual(
            utility.decIncSpeedLerp(1.0, 2.0, 1.0),
            2
        )


if __name__ == '__main__':
    unittest.main()



from collections import deque
from collections.abc import Callable

import pygame

from . import display
from .saveable import Saveable
from . import utility


class Anim(Saveable):
    __slots__ = (
        'func',
        'time',
        'pos',
        'sound',
        'positional_sound',
        'callback',
    )

    def __init__(self, func: str, time: int, x_or_pair, y=None,
                 sound: pygame.mixer.Sound = None, positional_sound: bool = False,
                 callback: Callable[[], None] = None):
        self.func = func
        self.time = time
        self.pos = pygame.math.Vector2(x_or_pair, y)
        self.sound = sound
        self.positional_sound = positional_sound
        self.callback = callback

    def save(self):
        # no sound right now, sorry
        # if we need it, either start passing sounds as paths
        # or don't save when there are pending Anims
        return self.func, self.time, self.pos

    @classmethod
    def load(cls, save_data):
        return cls(*save_data)


class AnimSprite(pygame.sprite.DirtySprite, Saveable):
    Binary = 'Binary'
    Lerp = 'LERP'
    IncSpeed = 'INC'
    DecSpeed = 'DEC'
    IncDecSpeed = 'INC_DEC'
    DecIncSpeed = 'DEC_INC'
    funcDict = {
        Binary: utility.binary,
        Lerp: utility.lerp,
        IncSpeed: utility.incSpeedLerp,
        DecSpeed: utility.decSpeedLerp,
        IncDecSpeed: utility.incDecSpeedLerp,
        DecIncSpeed: utility.decIncSpeedLerp,
    }

    @classmethod
    def toFunc(cls, func):
        return cls.funcDict.get(func, utility.lerp)

    __slots__ = (
        'anims',
        'last_pos',
        'time',
        'positional_sound',
        'sound_channel',
    )

    def __init__(self):
        super().__init__()
        # dirty = 2 : always draw
        self.dirty = 2
        self.anims = deque()
        self.last_pos = None
        self.time = 0
        self.positional_sound = False
        self.sound_channel = None

    def save(self):
        return {
            'rect_topleft': self.rect.topleft,
            'anims': self.anims,
            'last_pos': self.last_pos,
            'time': self.time,
        }

    @classmethod
    def load(cls, save_data):
        new_obj = cls()
        new_obj.rect.topleft = save_data['rect_topleft']
        new_obj.anims = save_data['anims']
        new_obj.last_pos = save_data['last_pos']
        new_obj.time = save_data['time']
        return new_obj

    def stillAnimating(self):
        if self.anims:
            return True
        return False

    def update(self, *args):
        if self.last_pos is None:
            self.last_pos = self.rect.center
        # adding dt
        self.time += args[0]
        while self.anims and self.time >= self.anims[0].time:
            done_anim = self.anims.popleft()
            self.time -= done_anim.time
            self.rect.center = done_anim.pos
            self.last_pos = self.rect.center
            if done_anim.sound:
                self.positional_sound = done_anim.positional_sound
                channel = done_anim.sound.play()
                if self.positional_sound:
                    self.sound_channel = channel
            if done_anim.callback:
                done_anim.callback()
        if self.anims:
            current_anim = self.anims[0]
            func = self.toFunc(current_anim.func)
            self.rect.center = func(
                self.last_pos,
                current_anim.pos,
                self.time / current_anim.time
            )
        else:
            self.last_pos = None
            self.time = 0
        if self.positional_sound:
            if self.sound_channel.get_busy():
                channel_l, channel_r = display.getPositionalChannelMix(self.rect.centerx)
                self.sound_channel.set_volume(channel_l, channel_r)
            else:
                self.positional_sound = False
                self.sound_channel = None

    def addPosAbs(self, func: str, time: int, x_or_pair, y=None,
                  sound: pygame.mixer.Sound = None, positional_sound: bool = False,
                  callback: Callable[[], None] = None):
        self.anims.append(
            Anim(func, time, x_or_pair, y, sound, positional_sound, callback)
        )

    def addPosRel(self, func: str, time: int, x_or_pair, y=None,
                  sound: pygame.mixer.Sound = None, positional_sound: bool = False,
                  callback: Callable[[], None] = None):
        new_pos = pygame.math.Vector2(x_or_pair, y)
        if self.anims:
            new_pos += self.anims[-1].pos
        else:
            new_pos += self.rect.center
        self.addPosAbs(func, time, new_pos, sound=sound, positional_sound=positional_sound, callback=callback)

    def addWait(self, time: int,
                sound: pygame.mixer.Sound = None, positional_sound: bool = False,
                callback: Callable[[], None] = None):
        self.addPosRel(AnimSprite.Binary, time, 0, 0, sound, positional_sound, callback)



import abc


class Saveable(abc.ABC):
    @abc.abstractmethod
    def save(self):
        """Return a serializable object representing all the information that should be saved from this object."""
        raise NotImplementedError(
            type(self).__name__ + ".saveMode(self)"
        )

    @classmethod
    @abc.abstractmethod
    def load(cls, save_data):
        """Take in an object (the result of a call to save()), and return an instance of this object."""
        raise NotImplementedError(
            cls.__name__ + ".loadMode(cls, saveData)"
        )



import sys
import os
import json
from types import ModuleType
from collections import deque

import pygame.math

from . import game
from .saveable import Saveable


_KEY_COLLECTION = 'COLLECTION'
_COLLECTION_DEQUE = 'DEQUE'
_COLLECTION_SET = 'SET'
_COLLECTION_VECTOR2 = 'VECTOR2'
_COLLECTION_VECTOR3 = 'VECTOR3'
_KEY_ELEMENTS = 'ELEMENTS'
_KEY_MAXLEN = 'MAXLEN'
_KEY_MODULE = 'MODULE'
_KEY_CLASS = 'CLASS'
_KEY_SAVEABLE = 'SAVEABLE'
_SAVE_EXT = '.sav'
_save_directory: str | None = None
_mode_module: ModuleType


def init(save_directory: str, mode_module: ModuleType):
    global _save_directory
    global _mode_module
    if _save_directory:
        raise RuntimeError("error: _save_directory is already set")
    _save_directory = save_directory
    _mode_module = mode_module


class _SaveableJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, deque):
            return {
                _KEY_COLLECTION: _COLLECTION_DEQUE,
                _KEY_ELEMENTS: list(o),
                _KEY_MAXLEN: o.maxlen,
            }
        elif isinstance(o, set):
            return {
                _KEY_COLLECTION: _COLLECTION_SET,
                _KEY_ELEMENTS: list(o),
            }
        elif isinstance(o, pygame.math.Vector2):
            return {
                _KEY_COLLECTION: _COLLECTION_VECTOR2,
                _KEY_ELEMENTS: list(o),
            }
        elif isinstance(o, pygame.math.Vector3):
            return {
                _KEY_COLLECTION: _COLLECTION_VECTOR3,
                _KEY_ELEMENTS: list(o),
            }
        elif isinstance(o, type):
            return {
                _KEY_MODULE: o.__module__,
                _KEY_CLASS: o.__qualname__,
            }
        elif isinstance(o, Saveable):
            return {
                _KEY_MODULE: type(o).__module__,
                _KEY_CLASS: type(o).__qualname__,
                _KEY_SAVEABLE: o.save(),
            }
        return super().default(o)


def _getClass(dct: dict):
    attr = sys.modules[dct[_KEY_MODULE]]
    for name in dct[_KEY_CLASS].split('.'):
        attr = getattr(attr, name)
    return attr


def _decodeSaveable(dct: dict):
    if _KEY_COLLECTION in dct:
        if dct[_KEY_COLLECTION] == _COLLECTION_DEQUE:
            return deque(dct[_KEY_ELEMENTS], dct[_KEY_MAXLEN])
        elif dct[_KEY_COLLECTION] == _COLLECTION_SET:
            return set(dct[_KEY_ELEMENTS])
        elif dct[_KEY_COLLECTION] == _COLLECTION_VECTOR2:
            return pygame.math.Vector2(dct[_KEY_ELEMENTS])
        elif dct[_KEY_COLLECTION] == _COLLECTION_VECTOR3:
            return pygame.math.Vector3(dct[_KEY_ELEMENTS])
    elif {_KEY_MODULE, _KEY_CLASS} == dct.keys():
        return _getClass(dct)
    elif {_KEY_MODULE, _KEY_CLASS, _KEY_SAVEABLE} == dct.keys():
        saveable_class = _getClass(dct)
        return saveable_class.load(dct[_KEY_SAVEABLE])
    return dct


class Save(object):
    __slots__ = (
        'save_name',
        '_mode_name',
        '_mode_data',
        '_shared_data',
    )

    def __init__(self, save_name: str, mode_name: str, mode_data, shared_data):
        self.save_name = save_name
        self._mode_name = mode_name
        self._mode_data = mode_data
        self._shared_data = shared_data

    @staticmethod
    def willOverwrite(save_name: str):
        return os.path.exists(
            Save._getFilePathFromFileName(save_name + _SAVE_EXT)
        )

    @staticmethod
    def _getSaveFiles():
        if not os.path.isdir(_save_directory):
            return ()
        return (
            file_name
            for file_name
            in os.listdir(_save_directory)
            if os.path.isfile(
                Save._getFilePathFromFileName(file_name)
            )
            and file_name.endswith(_SAVE_EXT)
        )

    @staticmethod
    def _getFilePathFromFileName(file_name):
        return os.path.join(_save_directory, file_name)

    def _getFilePath(self):
        return self._getFilePathFromFileName(self.save_name + _SAVE_EXT)

    @classmethod
    def getAllFromFiles(cls):
        return sorted(
            (
                save
                for save
                in (
                    cls._getFromFile(file)
                    for file
                    in cls._getSaveFiles()
                )
                if save
            ),
            key=lambda s: (s.save_name.lower(), s.save_name)
        )

    @classmethod
    def _getFromFile(cls, file_name: str):
        file_path = cls._getFilePathFromFileName(file_name)
        try:
            with open(file_path, 'r') as file:
                save_object = json.load(file, object_hook=_decodeSaveable)
                return cls(
                    file_name[:-len(_SAVE_EXT)],
                    save_object['mode_name'],
                    save_object['mode_data'],
                    save_object['shared_data']
                )
        except (IOError, json.decoder.JSONDecodeError):
            return False

    @classmethod
    def getFromMode(cls, save_name: str, from_mode: Saveable):
        return cls(
            save_name,
            type(from_mode).__name__,
            from_mode.save(),
            game.getGame().state.save()
        )

    def save(self):
        try:
            os.mkdir(_save_directory)
        except FileExistsError:
            pass
        save_object = {
            'mode_name': self._mode_name,
            'mode_data': self._mode_data,
            'shared_data': self._shared_data,
        }
        file_path = self._getFilePath()
        try:
            with open(file_path, 'w') as file:
                json.dump(save_object, file, cls=_SaveableJSONEncoder)
            return True
        except IOError:
            return False

    def load(self):
        game.getGame().state = game.getGame().state_cls.load(self._shared_data)
        mode_cls = getattr(_mode_module, self._mode_name)
        new_mode = mode_cls.load(self._mode_data)
        return new_mode

    def delete(self):
        file_path = self._getFilePath()
        os.remove(file_path)



import unittest

import pygame

import jovialengine.display as display


class TestDisplay(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        display.upscale = 3
        display.screen_size = (320, 240)

    @staticmethod
    def setUpFullScreen():
        display.is_fullscreen = True
        display._fullscreen_offset = [50, 100]

    @staticmethod
    def setUpWindowed():
        display.is_fullscreen = False
        display._fullscreen_offset = None

    def test_scaleMouseInput_fullscreen_button(self):
        # Arrange
        self.setUpFullScreen()
        event_dict = {
            'pos': (400, 800),
            'button': 1,
        }
        event = pygame.event.Event(pygame.MOUSEBUTTONDOWN, event_dict)
        # Act
        result_event = display.scaleMouseInput(event)
        # Assert
        self.assertEqual(result_event.pos, (116, 233))
        self.assertEqual(result_event.button, 1)

    def test_scaleMouseInput_fullscreen_motion(self):
        # Arrange
        self.setUpFullScreen()
        event_dict = {
            'pos': (400, 800),
            'rel': (100, 200),
            'buttons': (1,),
        }
        event = pygame.event.Event(pygame.MOUSEMOTION, event_dict)
        # Act
        result_event = display.scaleMouseInput(event)
        # Assert
        self.assertEqual(result_event.pos, (116, 233))
        self.assertEqual(result_event.rel, (33, 66))
        self.assertEqual(result_event.buttons, (1,))

    def test_scaleMouseInput_windowed_button(self):
        # Arrange
        self.setUpWindowed()
        event_dict = {
            'pos': (400, 800),
            'button': 1,
        }
        event = pygame.event.Event(pygame.MOUSEBUTTONDOWN, event_dict)
        # Act
        result_event = display.scaleMouseInput(event)
        # Assert
        self.assertEqual(result_event.pos, (133, 266))
        self.assertEqual(result_event.button, 1)

    def test_scaleMouseInput_windowed_motion(self):
        # Arrange
        self.setUpWindowed()
        event_dict = {
            'pos': (400, 800),
            'rel': (100, 200),
            'buttons': (1,),
        }
        event = pygame.event.Event(pygame.MOUSEMOTION, event_dict)
        # Act
        result_event = display.scaleMouseInput(event)
        # Assert
        self.assertEqual(result_event.pos, (133, 266))
        self.assertEqual(result_event.rel, (33, 66))
        self.assertEqual(result_event.buttons, (1,))

    def test_isInScreen_true(self):
        # Arrange
        self.setUpWindowed()
        # Act
        result = display.isInScreen((2, 3))
        # Assert
        self.assertTrue(result)

    def test_isInScreen_false(self):
        # Arrange
        self.setUpWindowed()
        # Act
        result = display.isInScreen((-2, 3))
        # Assert
        self.assertFalse(result)

    def test_getPositionalChannelMix(self):
        # Arrange
        self.setUpWindowed()
        # Act
        result = display.getPositionalChannelMix(90)
        # Assert
        self.assertEqual(result, (0.9231914344987546, 0.5420440747442257))


if __name__ == '__main__':
    unittest.main()
